/* 
 * VirtualBox Control Service
 * 
 * Copyright © 2010 vu1tur.eu.org
 * 
 * This file is part of VirtualBox Control Service
 * 
 * VirtualBox Control Service is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 of 
 * the License, or (at your option) any later version.
 * 
 * VirtualBox Control Service is distributed in the hope that it will 
 * be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "vboxctrl.h"
#include "vboxipc.h"
#include <psapi.h>
#include <tchar.h>

SERVICE_STATUS gSvcStatus;
SERVICE_STATUS_HANDLE gSvcStatusHandle;

char szModule[MAX_PATH];
BOOL fConfigured;

IVirtualBox *virtualBox;
ISession *session;
IConsole *console;
IEventSource *eventSource;
IEventListener *eventListener;
SAFEARRAY *eventTypes;
HANDLE hServiceThreadEvents[_EVENT_COUNT];
PROCESS_INFORMATION pi_vboxsvc;
HRESULT rcVMStop;

BSTR machineId = NULL;
char szvmName[255];
char bLogLevel = LOGLEVEL_DEFAULT;
char szLogFile[MAX_PATH];
char szVBoxInstallPath[MAX_PATH];
char szVBoxUserHome[MAX_PATH];
char szServiceName[255];
char szSvcName[255];
char szCmdShutdown[64];
char szPriority[64];
char szIniFile[MAX_PATH] = "";
DWORD dwMaxWaitTimeout;
DWORD PRIORITY_CLASS = DEFAULT_PRIORITY_CLASS;
DWORD dwShutdownAction;
BOOL bPreShutdownAvailable = false;
BOOL bInitializing = false;

int main(int argc, char* argv[])
{
	fConfigured = ReadIniFile(true);
	setenv();
	getver();
	if (argc>1)
	{
		if (_stricmp(argv[1], "-i") == 0)
		{
			if (!fConfigured) 
			{
				printf("\nConfiguration is not complete.\nCheck %s file.\n", szLogFile);
				return 1;
			}
			if (InstallService() == SVC_SUCCESS) 
			{
				printf("\n%s (%s) service installed successfully.\n", szServiceName, szSvcName);
				Log(1, "Service %s (%s) installed successfully.", szServiceName, szSvcName);
			} else 
				return 2;
		}
		else if (_stricmp(argv[1], "-u") == 0)
		{
			if (!fConfigured)
			{
				printf("\nConfiguration is not complete.\nCheck %s file.\n", szLogFile);
				return 1;
			}
			int ret = DeleteService();
			if (ret == SVC_SUCCESS)
			{
				printf("\n%s (%s) service uninstalled successfully.\n", szServiceName, szSvcName);
				Log(1, "Service %s (%s) uninstalled successfully.", szServiceName, szSvcName);
			} 
			else if (ret == SVC_RUNNING) 
			{
				printf("\n%s (%s) service is still running and will be uninstalled when stopped.\n", szServiceName, szSvcName);
				Log(1, "Service %s (%s) is still running and will be uninstalled when stopped.", szServiceName, szSvcName);
				return 2;
			}
		}
		else
		{
            DWORD tmp;
            UINT length;
            DWORD versionSize = GetFileVersionInfoSize(szModule, &tmp);
            VS_FIXEDFILEINFO *vFixInfo;
            if (versionSize > 0)
            {
                void *versionInfo = malloc(versionSize);
                if (versionInfo != NULL)
                {
                    GetFileVersionInfo(szModule, 0, versionSize, versionInfo);
                    VerQueryValue(versionInfo, "\\", (void**)&vFixInfo, &length);
                }
            
				printf("\nVirtualBox Control Service v%d.%d.%d%s\n\nUsage:\n\t %s -i\t\tInstall %s service\n\t %s -u\t\tUninstall %s service\n", 
					   HIWORD(vFixInfo->dwProductVersionMS), 
					   LOWORD(vFixInfo->dwProductVersionMS), 
					   HIWORD(vFixInfo->dwProductVersionLS), 
#ifdef __x86_64__
					   " (x64)",
#else
					   "",
#endif
                   szSvcName, szSvcName, szSvcName, szSvcName);
				if (versionInfo != NULL)
					free(versionInfo);                    
			}
		}
	} else {
			SERVICE_TABLE_ENTRY ste[]={{szSvcName, ServiceMain},{NULL,NULL}};
			StartServiceCtrlDispatcher(ste);
	}
	return 0;
}

BOOL ReadIniFile(BOOL bAll)
{
	GetModuleFileName(NULL, szModule, MAX_PATH);
	strcpy(szIniFile, szModule);
	szIniFile[strlen(szIniFile)-4] = 0;
	strcat(szIniFile, ".ini");

	ZeroMemory(szLogFile, MAX_PATH);
	strcat(szLogFile, szModule);
	szLogFile[strlen(szLogFile)-4] = 0;
	strcat(szLogFile, ".log");

	if (_access(szIniFile, 0) == -1)
	{
		Log(1, "Configuration file '%s' does not exist", szIniFile);
		return false;
	}

	bLogLevel = GetPrivateProfileInt("VBoxCtrl", "Loglevel", LOGLEVEL_DEFAULT, szIniFile);
	
    if (bAll)
    {
    	if (strlen(strrchr(szModule, '\\')+1) > 4)
    		strncpy(szSvcName, strrchr(szModule, '\\')+1, strlen(strrchr(szModule, '\\')+1) - 4);
    	else 
    		strcpy(szSvcName, VBOX_CONTROL_SERVICE_NAME);
    	strcrpl(szSvcName, ' ', '_');

    	GetPrivateProfileString("VBoxCtrl", "VboxInstallPath", "", szVBoxInstallPath, MAX_PATH, szIniFile);
    	if (strlen(szVBoxInstallPath) == 0)
    	{
    		Log(1, "ERROR: VboxInstallPath not configured");
    		return false;
    	}
    	strcrpl(szVBoxInstallPath, '/', '\\');
    	if (szVBoxInstallPath[strlen(szVBoxInstallPath)-1] == '\\')
    		szVBoxInstallPath[strlen(szVBoxInstallPath)-1] = 0;
	
    	GetPrivateProfileString("VBoxCtrl", "VboxUserHome", "", szVBoxUserHome, MAX_PATH, szIniFile);
    	if (strlen(szVBoxUserHome) == 0)
    	{
    		Log(1, "ERROR: VboxUserHome not configured");
    		return false;
    	}
    	strcrpl(szVBoxUserHome, '/', '\\');
    	if (szVBoxUserHome[strlen(szVBoxUserHome)-1] == '\\')
    		szVBoxUserHome[strlen(szVBoxUserHome)-1] = 0;
    		
		GetPrivateProfileString("VBoxCtrl", "VmName", "", szvmName, 255, szIniFile);
    	if (strlen(szvmName) == 0)
    	{
    		Log(1, "ERROR: No VmName configured");
    		return false;	
    	}
    	
        GetPrivateProfileString("VBoxCtrl", "ServiceName", NULL, szServiceName, 255, szIniFile);
        
      	GetPrivateProfileString("VBoxCtrl", "Priority", NULL, szPriority, 64, szIniFile);
    	if (_stricmp(szPriority, "idle") == 0)
    		PRIORITY_CLASS = IDLE_PRIORITY_CLASS;
    	else if (_stricmp(szPriority, "below normal") == 0)
    		PRIORITY_CLASS = BELOW_NORMAL_PRIORITY_CLASS;
    	else if (_stricmp(szPriority, "normal") == 0)
    		PRIORITY_CLASS = NORMAL_PRIORITY_CLASS;
    	else if (_stricmp(szPriority, "above normal") == 0)
    		PRIORITY_CLASS = ABOVE_NORMAL_PRIORITY_CLASS;
    	else if (_stricmp(szPriority, "high") == 0)
    		PRIORITY_CLASS = HIGH_PRIORITY_CLASS;
    	else
    		PRIORITY_CLASS = DEFAULT_PRIORITY_CLASS;

    }	
    
	GetPrivateProfileString("VBoxCtrl", "ShutdownOnceCommand", NULL, szCmdShutdown, 64, szIniFile);
	if (strlen(szCmdShutdown) == 0)
		GetPrivateProfileString("VBoxCtrl", "ShutdownCommand", SHUTDOWN_ACTION_DEFAULT_SZ, szCmdShutdown, 64, szIniFile);
	if (_stricmp(szCmdShutdown, "poweroff") == 0)
		dwShutdownAction = SHUTDOWN_ACTION_POWEROFF;
	else if (_stricmp(szCmdShutdown, "savestate") == 0)
		dwShutdownAction = SHUTDOWN_ACTION_SAVESTATE;
	else if (_stricmp(szCmdShutdown, "acpisleepbutton") == 0)
		dwShutdownAction = SHUTDOWN_ACTION_SLEEPBTN;
	else if (_stricmp(szCmdShutdown, "acpipowerbutton") == 0)
		dwShutdownAction = SHUTDOWN_ACTION_POWERBTN;
	else {
		strcpy(szCmdShutdown, SHUTDOWN_ACTION_DEFAULT_SZ);
		dwShutdownAction = SHUTDOWN_ACTION_DEFAULT;
	}
		
	dwMaxWaitTimeout = GetPrivateProfileInt("VBoxCtrl", "MaxWaitTimeout", MAX_WAIT_TIMEOUT, szIniFile);
		
	return true;
}

void WINAPI ServiceMain(DWORD argc, LPTSTR *argv)
{
	UNUSED(argc);
	UNUSED(argv);
	gSvcStatusHandle = RegisterServiceCtrlHandlerEx(szSvcName, ServiceCtrlHandler, NULL);
	if (gSvcStatusHandle == (SERVICE_STATUS_HANDLE)0)
		return;

	gSvcStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS;
	gSvcStatus.dwControlsAccepted = SERVICE_ACCEPT_SHUTDOWN | SERVICE_ACCEPT_STOP;
	if (bPreShutdownAvailable)
	   gSvcStatus.dwControlsAccepted |= SERVICE_ACCEPT_PRESHUTDOWN;
	gSvcStatus.dwWin32ExitCode = 0;
	gSvcStatus.dwCheckPoint = 0;
	gSvcStatus.dwWaitHint = 0;

	if (!ReadIniFile(true)) {
		gSvcStatus.dwCurrentState = SERVICE_STOPPED;
		SetServiceStatus(gSvcStatusHandle, &gSvcStatus);
		return;
	}
	
	DWORD tmp;
	UINT length;
	DWORD versionSize = GetFileVersionInfoSize(szModule, &tmp);
	VS_FIXEDFILEINFO *vFixInfo;
	if (versionSize > 0)
	{
		void *versionInfo = malloc(versionSize);
		if (versionInfo != NULL)
		{
			GetFileVersionInfo(szModule, 0, versionSize, versionInfo);
			VerQueryValue(versionInfo, "\\", (void**)&vFixInfo, &length);
			free(versionInfo);                    
		}
	}
	Log(3, "VirtualBox Control Service v%d.%d.%d.%d%s started", 
		HIWORD(vFixInfo->dwProductVersionMS),
		LOWORD(vFixInfo->dwProductVersionMS),
		HIWORD(vFixInfo->dwProductVersionLS),
		LOWORD(vFixInfo->dwProductVersionLS),
#ifdef __x86_64__
					" (x64)"
#else
					""
#endif		
		);
	
	ReportSvcStatus(SERVICE_START_PENDING, NO_ERROR, 0);
	if (start_vbox_svc())
		Sleep(900);
	_beginthread(ServiceThread, 0, NULL);
	_beginthread(VBoxIPCListener, 0, NULL);
}

BOOL start_vbox_svc()
{
	BOOL isVboxSvcRunning = FALSE;
	DWORD ProcessIds[16384], cbRet, dwSize;
	HANDLE hToken;
	OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken);
	GetTokenInformation(hToken, TokenUser, NULL, 0, &dwSize);
	TOKEN_USER *pTokenCurrentUser = (TOKEN_USER*) malloc(dwSize);
	GetTokenInformation(hToken, TokenUser, (void*)pTokenCurrentUser, dwSize, &dwSize);
	CloseHandle(hToken);
	
	if (EnumProcesses(ProcessIds, sizeof(ProcessIds), &cbRet)) {
		for (int i=cbRet/sizeof(DWORD);i--;) {
			if (ProcessIds[i] == 0)
				continue;
			TCHAR szProcessName[MAX_PATH] = TEXT("<>");
			HANDLE hProcess = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, ProcessIds[i]);
			if (hProcess == NULL)
				continue;
			HMODULE hMod;
			DWORD cbNeeded;
			if (EnumProcessModules(hProcess, &hMod, sizeof(hMod), &cbNeeded)) {
				GetModuleBaseName(hProcess, hMod, szProcessName, sizeof(szProcessName)/sizeof(TCHAR));
				if (_tcscmp(szProcessName, _T("VBoxSVC.exe")) == 0) {
					if (OpenProcessToken(hProcess, TOKEN_QUERY, &hToken)) {
						GetTokenInformation(hToken, TokenUser, NULL, 0, &dwSize);
						TOKEN_USER *pTokenUser = (TOKEN_USER*) malloc(dwSize);
						if (GetTokenInformation(hToken, TokenUser, (void*)pTokenUser, dwSize, &dwSize)) {
							if (EqualSid(pTokenUser->User.Sid, pTokenCurrentUser->User.Sid)) {
								isVboxSvcRunning = TRUE;
								break;
							}
						}
						if (pTokenUser)
							free(pTokenUser);
						CloseHandle(hToken);
					}
				}
			}
			CloseHandle(hProcess);
		}
	}
	if (pTokenCurrentUser)
		free(pTokenCurrentUser);

	if (isVboxSvcRunning)
		return FALSE;
		
	char szCmd[256] = "";
	sprintf(szCmd, "%s\\VBoxSVC.exe", szVBoxInstallPath);
	STARTUPINFO si;
	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	ZeroMemory(&pi_vboxsvc, sizeof(pi_vboxsvc));

	if (CreateProcess(NULL, szCmd, NULL, NULL, false, CREATE_NO_WINDOW | PRIORITY_CLASS, NULL, NULL, &si, &pi_vboxsvc)) {
		Log(3, "Started VBoxSVC[%d]", pi_vboxsvc.dwProcessId);
	} else {
		Log(2, "Could not manually start VBoxSVC: %s", GetLastError());
	}
	
	return TRUE;
}

void ServiceThread(void *args)
{
    UNUSED(args);
    DWORD svcExitCode = NO_ERROR;
    BOOL running = TRUE;
	hServiceThreadEvents[_EVENT_STOPVM] = CreateEvent(NULL, FALSE, FALSE, NULL);
	hServiceThreadEvents[_EVENT_STOPSVC] = CreateEvent(NULL, FALSE, FALSE, NULL);
	hServiceThreadEvents[_EVENT_ABORTVM] = CreateEvent(NULL, FALSE, FALSE, NULL);
	InitializeVirtualBoxInterface();
	ReportSvcStatus(SERVICE_START_PENDING, NO_ERROR, 0);
	if (StartVirtualBoxMachine(szvmName))
	{
    	while(running)
    	{	
            switch (WaitForMultipleObjects(_EVENT_COUNT, hServiceThreadEvents, FALSE, INFINITE))
            {
                case WAIT_OBJECT_0 + _EVENT_STOPVM:
                    _beginthread(ProgressThread, 0, NULL);
                    StopVirtualBoxMachine(szvmName);
                    break;
                case WAIT_OBJECT_0 + _EVENT_STOPSVC:
					Log(2, "VM '%s' stopped (%s)", szvmName, szCmdShutdown);
               		running = FALSE;
                    break;
				case WAIT_OBJECT_0 + _EVENT_ABORTVM:
					Log(1, "VM '%s' was aborted", szvmName);
					running = FALSE;
					break;
                default:
                    break;
            }
    	}
    }
    else
    {
        gSvcStatus.dwServiceSpecificExitCode = 0x80800001;
        svcExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
    }
	Sleep(500);
	VBoxIPCShutdown();
    CloseVirtualBoxInterface();
	ReportSvcStatus(SERVICE_STOPPED, svcExitCode, 0);
}

void ProgressThread(void *args)
{
	UNUSED(args);
    while (gSvcStatus.dwCurrentState == SERVICE_STOP_PENDING)
    {
        ReportSvcStatus(SERVICE_STOP_PENDING, NO_ERROR, dwMaxWaitTimeout);
        Sleep(1000);
    }
}

void ListenerThread(void *args)
{
	UNUSED(args);
	HRESULT rc;
	IEvent* ev;
	IMachineStateChangedEvent* mev;
	MachineState state;
	VBoxEventType type;
	BSTR mId = NULL;
	while (virtualBox != NULL) {
		rc = eventSource->GetEvent(eventListener, 500, &ev);
		if (SUCCEEDED(rc)) {
			if (ev != NULL) {
				ev->get_Type(&type);
				if (type == VBoxEventType_OnMachineStateChanged) {
					rc = ev->QueryInterface(IID_IMachineStateChangedEvent, (void**)&mev);
					if (SUCCEEDED(rc)) {
						mev->get_MachineId(&mId);
						if (wcscmp(mId, machineId) == 0) {
							rc = mev->get_State(&state);
							if (SUCCEEDED(rc))
								switch (state) {
									case MachineState_Running:
										ReportSvcStatus(SERVICE_RUNNING, NO_ERROR, 0);
										break;
									case MachineState_Aborted:
										SetEvent(hServiceThreadEvents[_EVENT_ABORTVM]);
										break;
									case MachineState_PoweredOff:
									case MachineState_Saved:
										SetEvent(hServiceThreadEvents[_EVENT_STOPSVC]);
										break;
									case MachineState_Saving:
									case MachineState_Stopping:
										ReportSvcStatus(SERVICE_STOP_PENDING, NO_ERROR, 0);
									default:
										break;	
							}
							VBoxIPCC_Reload();
						}
						SysFreeString(mId);
					}
				}
				eventSource->EventProcessed(eventListener, ev);
				SAFE_RELEASE(mev);
				SAFE_RELEASE(ev);
			}
		}
	}
}

void setenv()
{
	char _szVBoxInstallPath[MAX_PATH];
	sprintf(_szVBoxInstallPath, "%s\\", szVBoxInstallPath);
	SetEnvironmentVariable("VBOX_INSTALL_PATH", _szVBoxInstallPath);
	SetEnvironmentVariable("VBOX_USER_HOME", szVBoxUserHome);
}

void getver()
{
    OSVERSIONINFOEX osvi;
    DWORDLONG dwConditionMask = 0;
    ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));
    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
    osvi.dwMajorVersion = 6;
    osvi.dwMinorVersion = 0;
    osvi.wServicePackMajor = 0;
    osvi.wServicePackMinor = 0;
    VER_SET_CONDITION(dwConditionMask, VER_MAJORVERSION, VER_GREATER_EQUAL);
    VER_SET_CONDITION(dwConditionMask, VER_MINORVERSION, VER_GREATER_EQUAL);
    VER_SET_CONDITION(dwConditionMask, VER_SERVICEPACKMAJOR, VER_GREATER_EQUAL);
    VER_SET_CONDITION(dwConditionMask, VER_SERVICEPACKMINOR, VER_GREATER_EQUAL);

    bPreShutdownAvailable = 
        VerifyVersionInfo(&osvi, VER_MAJORVERSION | VER_MINORVERSION | 
                          VER_SERVICEPACKMAJOR | VER_SERVICEPACKMINOR, 
                          dwConditionMask);
}

DWORD WINAPI ServiceCtrlHandler(DWORD Opcode, DWORD dwEventType, void* lpEventData, void* lpContext)
{
	UNUSED(dwEventType);
	UNUSED(lpEventData);
	UNUSED(lpContext);
	switch(Opcode)
	{
		case SERVICE_CONTROL_STOP:
        case SERVICE_CONTROL_PRESHUTDOWN:
		case SERVICE_CONTROL_SHUTDOWN:
            ReadIniFile(false);
			ReportSvcStatus(SERVICE_STOP_PENDING, NO_ERROR, dwMaxWaitTimeout);
#ifdef DEBUG
			Log(3, "%s", Opcode == SERVICE_CONTROL_STOP ? "SERVICE_CONTROL_STOP" : 
                (Opcode == SERVICE_CONTROL_SHUTDOWN ? "SERVICE_CONTROL_SHUTDOWN" : 
                "SERVICE_CONTROL_PRESHUTDOWN"));
#endif
			SetEvent(hServiceThreadEvents[_EVENT_STOPVM]);
			return NO_ERROR;

		case SERVICE_CONTROL_INTERROGATE:
			return NO_ERROR;

		default:
			break;
	}
	return ERROR_CALL_NOT_IMPLEMENTED;
}

void ReportSvcStatus(DWORD dwCurrentState, DWORD dwWin32ExitCode, DWORD dwWaitHint)
{
    static DWORD dwCheckPoint = 1;
    gSvcStatus.dwCurrentState = dwCurrentState;
    gSvcStatus.dwWin32ExitCode = dwWin32ExitCode;
    gSvcStatus.dwWaitHint = dwWaitHint;
    if ((dwCurrentState == SERVICE_RUNNING) ||
        (dwCurrentState == SERVICE_STOPPED))
        gSvcStatus.dwCheckPoint = 0;
    else gSvcStatus.dwCheckPoint = dwCheckPoint++;

    if (!SetServiceStatus(gSvcStatusHandle, &gSvcStatus))
    	Log(1, "ERROR settings service status: 0x%08x", GetLastError());
}

int InstallService()
{
	SC_HANDLE hSCManager, hService;
	GetModuleFileName(NULL,szModule,MAX_PATH);
	hSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if (hSCManager == NULL)
		return SVC_ERROR;
	LPCTSTR lpszBinaryPathName = szModule;
	hService = CreateService(hSCManager, szSvcName, szServiceName, SERVICE_ALL_ACCESS,
		SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS,
		SERVICE_AUTO_START,
		SERVICE_ERROR_NORMAL,
		lpszBinaryPathName,
		NULL, NULL, NULL, NULL, NULL);
	if (hService == NULL)
	{
		char errmsg[255] = "";
		switch(GetLastError())
		{
			case ERROR_SERVICE_EXISTS:
				strcpy(errmsg, "service already exists");
				break;
			case ERROR_DUPLICATE_SERVICE_NAME:
				strcpy(errmsg, "duplicate service display name");
				break;
			case ERROR_ACCESS_DENIED:
				strcpy(errmsg, "access denied");
				break;
			case ERROR_INVALID_NAME:
				strcpy(errmsg, "invalid name");
				break;
		}
		Log(1, "ERROR installing %s (%s) service: %s", szServiceName, szSvcName, errmsg);
		printf("\nERROR installing %s (%s) service: %s\n", szServiceName, szSvcName, errmsg);
		CloseServiceHandle(hSCManager);
		return SVC_ERROR;		
	}

	SERVICE_DESCRIPTION svcDescr;
	svcDescr.lpDescription = "Provides automatic start and shutdown of VirtualBox machine.";
	ChangeServiceConfig2(hService, SERVICE_CONFIG_DESCRIPTION, &svcDescr);

	if (bPreShutdownAvailable) 
    {
        SERVICE_PRESHUTDOWN_INFO svcPreShutdownInfo;
    	svcPreShutdownInfo.dwPreshutdownTimeout = MAX_WAIT_TIMEOUT;
    	ChangeServiceConfig2(hService, SERVICE_CONFIG_PRESHUTDOWN_INFO, &svcPreShutdownInfo);
    }
	
	SetServicePermissions(hService);
	
	CloseServiceHandle(hService);
	CloseServiceHandle(hSCManager);
	return SVC_SUCCESS;
}

int DeleteService()
{
	int ret = SVC_SUCCESS;
	SC_HANDLE hSCManager, hService;
	hSCManager = OpenSCManager(NULL, NULL, GENERIC_READ);
	if (hSCManager == NULL)
		return SVC_ERROR;
	hService = OpenService(hSCManager, szSvcName, SERVICE_ALL_ACCESS);
	if (hService == NULL)
	{
		Log(1, "ERROR uninstalling %s (%s) service: service not found", szServiceName, szSvcName);
		printf("\nERROR uninstalling %s (%s) service: service not found\n", szServiceName, szSvcName);
		CloseServiceHandle(hSCManager);
		return SVC_ERROR;
	}
	
	SERVICE_STATUS svcStatus;
	QueryServiceStatus(hService, &svcStatus);
	if (svcStatus.dwCurrentState != SERVICE_STOPPED)
		ret = SVC_RUNNING;
	
	if (DeleteService(hService) == 0)
	{
		char errmsg[255];
		switch(GetLastError())
		{
			case ERROR_ACCESS_DENIED:
				strcpy(errmsg, "access denied");
				break;
			case ERROR_SERVICE_MARKED_FOR_DELETE:
				strcpy(errmsg, "service already marked for deletion");
				break;
		}
		Log(1, "ERROR uninstalling %s (%s) service: %s", szServiceName, szSvcName, errmsg);
		printf("\nERROR uninstalling %s (%s) service: %s\n", szServiceName, szSvcName, errmsg);
		CloseServiceHandle(hService);
		CloseServiceHandle(hSCManager);
		return SVC_ERROR;
	}
	CloseServiceHandle(hService);
	CloseServiceHandle(hSCManager);
	return ret;
}

void SetServicePermissions(SC_HANDLE hService)
{
	EXPLICIT_ACCESS ea;
	SECURITY_DESCRIPTOR	sd;
	PSECURITY_DESCRIPTOR psd = { 0 };
	PACL pacl = NULL;
	PACL pnewacl = NULL;
	BOOL bDaclPresent = FALSE;
	BOOL bDaclDefaulted = FALSE;
	DWORD dwSize = 0;
	DWORD dwBytesNeeded = 0;
	DWORD lasterr;
	SID_IDENTIFIER_AUTHORITY SIDAuth = { SECURITY_NT_AUTHORITY };
	PSID InteractiveSID;
	int err = 1;
	
	while (1) {
		if (!QueryServiceObjectSecurity(hService, DACL_SECURITY_INFORMATION, psd, 0, &dwBytesNeeded))
		{
			lasterr = GetLastError();
			if (lasterr == ERROR_INSUFFICIENT_BUFFER)
			{
				dwSize = dwBytesNeeded;
				psd = (PSECURITY_DESCRIPTOR)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwSize);
				if (psd == NULL)
					break;
				if (!QueryServiceObjectSecurity(hService, DACL_SECURITY_INFORMATION, psd, dwSize, &dwBytesNeeded))
					break;
			} else {
				Log(1, "ERROR retrieving security descriptor of the service (0x%08x)", lasterr);
				break;
			}
		}
		if (!GetSecurityDescriptorDacl(psd, &bDaclPresent, &pacl, &bDaclDefaulted)) {
			Log(1, "ERROR retrieving DACL from security descriptor (0x%08x)", GetLastError());
			break;
		}

		if (!AllocateAndInitializeSid(&SIDAuth, 1, SECURITY_INTERACTIVE_RID, 0, 0, 0, 0, 0, 0, 0, &InteractiveSID)) {
			Log(1, "ERROR initializing SID for interactive user (0x%08x)", GetLastError());
			break;
		}
		
		ZeroMemory(&ea, sizeof(EXPLICIT_ACCESS));
		ea.grfAccessMode = GRANT_ACCESS;
		ea.grfAccessPermissions = SERVICE_START | SERVICE_STOP | SERVICE_CHANGE_CONFIG | GENERIC_READ;
		ea.grfInheritance = NO_INHERITANCE;
		ea.Trustee.TrusteeForm = TRUSTEE_IS_SID;
		ea.Trustee.TrusteeType = TRUSTEE_IS_WELL_KNOWN_GROUP;
		ea.Trustee.ptstrName = (LPTSTR)InteractiveSID;

		lasterr = SetEntriesInAcl(1, &ea, pacl, &pnewacl);
		if (lasterr != ERROR_SUCCESS) {
			Log(1, "ERROR creating ACL (0x%08x, 0x%08x)", lasterr, GetLastError());
			break;
		}

		if (!InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION)) {
			Log(1, "ERROR initializing new security descriptor (0x%08x)", GetLastError());
			break;
		}

		if (!SetSecurityDescriptorDacl(&sd, TRUE, pnewacl, FALSE)) {
			Log(1, "ERROR replacing DACL (0x%08x)", GetLastError());
			break;
		}

		if (!SetServiceObjectSecurity(hService, DACL_SECURITY_INFORMATION, &sd)) {
			Log(1, "ERROR setting security descriptor of a service (0x%08x)", GetLastError());
			break;
		}
		
		err = 0;
		break;
	}
	if (pnewacl != NULL)
		LocalFree((HLOCAL)pnewacl);
	if (psd != NULL)
		HeapFree(GetProcessHeap(), 0, (LPVOID)psd);
	if (InteractiveSID != NULL)
		FreeSid(InteractiveSID);
	if (err != 0)
		Log(1, "ERROR updating service DACL");
}

void CreateEventHandler()
{
	HRESULT rc;
	rc = virtualBox->get_EventSource(&eventSource);
	if (!SUCCEEDED(rc)) {
		Log(1, "Error getting event source interface (0x%08x)", rc);
		return;
	}
	rc = eventSource->CreateListener((IEventListener**)&eventListener);
	if (!SUCCEEDED(rc)) {
		Log(1, "Error creating event listener (0x%08x)", rc);
		return;
	}
	VBoxEventType *t;
	SAFEARRAYBOUND bound = {1,0};
	eventTypes = SafeArrayCreate(VT_I4, 1, &bound);
	rc = SafeArrayAccessData(eventTypes,(void**)&t);
	if (SUCCEEDED(rc)) {
		t[0] = VBoxEventType_OnMachineStateChanged;
		SafeArrayUnaccessData(eventTypes);
		rc = eventSource->RegisterListener(eventListener, eventTypes, false);
		if (!SUCCEEDED(rc)) {
			Log(1, "Error registering event listener (0x%08x)", rc);
			return;
		}
		_beginthread(ListenerThread, 0, NULL);
	}
	SafeArrayDestroy(eventTypes);
}

HRESULT InitializeVirtualBoxInterface()
{
	HRESULT rc = InitializeVirtualBoxInterface(&virtualBox);
	if (SUCCEEDED(rc))
		rc = OpenVirtualBoxSession(&session);
	return rc;
}

HRESULT InitializeVirtualBoxInterface(IVirtualBox **ppVirtualBox)
{
	if (bInitializing)
		return 0x77770004;
	bInitializing = TRUE;	
	HRESULT rc;
	BSTR virtualBoxVersion;
	BSTR homedir;
	IMachine *machine;
	wchar_t wcsVBoxUserHome[strlen(szVBoxUserHome)+1];
	mbstowcs(wcsVBoxUserHome, szVBoxUserHome, strlen(szVBoxUserHome)+1);	
	
	CoInitialize(NULL);
	Log(3, "Initializing VirtualBox interface");

	for (int i=3;i--;) {
		rc = CoCreateInstance(CLSID_VirtualBox, NULL, CLSCTX_LOCAL_SERVER, IID_IVirtualBox, (void**)ppVirtualBox);
		if (!SUCCEEDED(rc)) {
			Log(1, "ERROR creating VirtualBox instance (rc = 0x%08x)", (DWORD)rc);
			bInitializing = FALSE;
			return rc;
		}
		(*ppVirtualBox)->get_HomeFolder(&homedir);
		if (wcscmp(homedir, wcsVBoxUserHome) == 0)
			break;
		else {
			if (i > 0)
				SAFE_RELEASE((*ppVirtualBox)); 
			Sleep(5500);
		}
	}

	Log(3, "Using VBOX_USER_HOME: %S", homedir);
	SysFreeString(homedir);

	rc = (*ppVirtualBox)->get_Version(&virtualBoxVersion);
	if (SUCCEEDED(rc))
		Log(3, "VirtualBox Interface v%S initialized", virtualBoxVersion);
	else
		Log(1, "ERROR initializing VirtualBox interface");
	SysFreeString(virtualBoxVersion);

	if (GetVirtualBoxMachine(szvmName, &machine)) {
		rc = machine->get_Id(&machineId);
		if (!SUCCEEDED(rc))
			Log(1, "ERROR retrieving machine ID (0x%08x)", (DWORD)rc);
	}
	SAFE_RELEASE(machine);
	
	CreateEventHandler();

	bInitializing = FALSE;
	return rc;
}

HRESULT OpenVirtualBoxSession(ISession **ppSession)
{
	HRESULT rc = CoCreateInstance(CLSID_Session, NULL, CLSCTX_INPROC_SERVER, IID_ISession, (void**)ppSession);
	if (!SUCCEEDED(rc))
		Log(1, "ERROR creating session instance (0x%08x)", (DWORD)rc);
	return rc;
}

void CloseVirtualBoxInterface()
{
	SysFreeString(machineId);
	SAFE_RELEASE(session);
	CloseVirtualBoxInterface(&virtualBox);
}

void CloseVirtualBoxInterface(IVirtualBox **ppVirtualBox)
{
	Log(3, "Closing VirtualBox interface");
	
	if (eventSource) {
		eventSource->UnregisterListener(eventListener);
		SAFE_RELEASE(eventListener);
		SAFE_RELEASE(eventSource);
	}
	
	SAFE_RELEASE((*ppVirtualBox));
	CoUninitialize();
}

BSTR alloc_bstr(const char* str)
{
	int ulength = MultiByteToWideChar(CP_ACP, 0, str, -1, NULL, 0);
	BSTR bstr = SysAllocStringLen(NULL, ulength);
	MultiByteToWideChar(CP_ACP, 0, str, -1, bstr, ulength);
	return bstr;
}

BOOL GetVirtualBoxMachine(const char* vm, IMachine **machine)
{
	return GetVirtualBoxMachine(virtualBox, vm, machine);
}

BOOL GetVirtualBoxMachine(IVirtualBox *pVirtualBox, const char* vm, IMachine **machine)
{
    HRESULT rc;
    BSTR machineName = alloc_bstr(vm);
    rc = pVirtualBox->FindMachine(machineName, machine);
    if (FAILED(rc))
    {
        if (rc == (HRESULT) VBOX_E_OBJECT_NOT_FOUND)
            Log(1, "ERROR: Could not find registered VM '%s'", vm);
        else
            Log(1, "ERROR getting virtual machine interface (0x%08x)", (DWORD)rc);
    } 

    SysFreeString(machineName);
    return SUCCEEDED(rc);
}

BOOL StartVirtualBoxMachine(const char* vm)
{
    BOOL rv = true;
    DWORD sessionPid = 0;
    IMachine *machine = NULL;
	IProgress *progress = NULL;
	BSTR launchType = alloc_bstr("headless");
	
	char _tmp_filename[1024];
	sprintf(_tmp_filename, "%s\\VBoxHeadless.exe", szVBoxInstallPath);
	if (_access(_tmp_filename, 0) == -1) {
		Log(2, "ERROR: Can't find VBoxHeadless executable. Check your VboxInstallPath option in %s", szIniFile);
		return false;
	}

    if (GetVirtualBoxMachine(vm, &machine))
    {
        MachineState machineState;
        HRESULT rc = machine->get_State(&machineState);
        if (SUCCEEDED(rc))
        {
            switch (machineState)
            {
                case MachineState_Starting:
                case MachineState_Restoring:
                case MachineState_Running:
                case MachineState_Paused:
                    machine->get_SessionPID(&sessionPid);
                    Log(2, "VM '%s' is already running (pid=%d), using existing session", vm, sessionPid);
					rc = machine->LockMachine(session, LockType_Shared);
					if (!SUCCEEDED(rc))
						Log(1, "Error getting a machine lock (0x%08x)", rc);
                    break;
                    
                case MachineState_Stopping:
                case MachineState_Saving:
                    Log(1, "ERROR starting VM '%s' because it is stopping", vm);
                    rv = false;
                    break;
                    
                default:
					Log(2, "Starting VM '%s'", vm);
					machine->LaunchVMProcess(session, launchType, NULL, &progress);
					if (progress) {
						rc = progress->WaitForCompletion(-1);
						if (!SUCCEEDED(rc)) {
							Log(1, "Error starting VM '%s' (0x%08x)", vm, rc);
							break;
						}
						Log(2, "VM '%s' started", vm);
					}
                    break;
            }
        }
        else {
            Log(1, "ERROR reading VM state (0x%08x)", (DWORD)rc);
            rv = false;
        }
        
        SAFE_RELEASE(machine);
     }
     else
        rv = false;
     
	 SysFreeString(launchType);
	 return rv;
}

void StopVirtualBoxMachine(const char* vm)
{
	Log(2, "Stopping (%s) VM '%s'", szCmdShutdown, vm);
	IMachine *machine = NULL;
	WritePrivateProfileString("VBoxCtrl", "ShutdownOnceCommand", NULL, szIniFile);
	
	if (GetVirtualBoxMachine(vm, &machine))
	{
		IProgress *progress = NULL;
		do
		{
			HRESULT rc = session->get_Console(&console);
			if (!SUCCEEDED(rc)) {
				Log(1, "ERROR getting console (0x%08x)", (DWORD)rc);
				break;
			}
			
			switch (dwShutdownAction)
			{
				case SHUTDOWN_ACTION_POWEROFF:
					rc = console->PowerDown(&progress);
					break;
				case SHUTDOWN_ACTION_SAVESTATE:
					rc = console->SaveState(&progress);
					break;
				case SHUTDOWN_ACTION_POWERBTN:
					rc = console->PowerButton();
					break;
				case SHUTDOWN_ACTION_SLEEPBTN:
					rc = console->SleepButton();
					break;
				default:
					rc = console->PowerDown(&progress);
			}
			if (!SUCCEEDED(rc))
				Log(1, "ERROR stopping VM '%s' (0x%08x)", vm, (DWORD)rc);
			if (progress) {
				rcVMStop = progress->WaitForCompletion(-1);
				if (!SUCCEEDED(rcVMStop))
					Log(1, "ERROR stopping (%s) VM '%' (0x%08x)", szCmdShutdown, vm, (DWORD)rcVMStop);
			} else if (SUCCEEDED(rc)) {
				Log(2, "VM '%s' was sent a stop signal (%s)", vm, szCmdShutdown);
			}
			session->UnlockMachine();
			SetEvent(hServiceThreadEvents[_EVENT_STOPSVC]);
		} while (0);
		
		SAFE_RELEASE(console);
		SAFE_RELEASE(progress);
		SAFE_RELEASE(machine);
	}
}

BOOL GetMachineVRDPEnabled()
{
	BOOL isEnabled = FALSE;
	IMachine *machine = NULL;
	IVRDEServer *vrdpServer = NULL;
	
	if (GetVirtualBoxMachine(virtualBox, szvmName, &machine)) {
		if (SUCCEEDED(machine->get_VRDEServer(&vrdpServer))) {
			vrdpServer->get_Enabled(&isEnabled);
			SAFE_RELEASE(vrdpServer);
		}
		SAFE_RELEASE(machine);
	}
	
	return isEnabled;
}

LONG GetMachineVRDPPort()
{
	IMachine *machine = NULL;
	LONG vrdpPort = 0;
	
	if (GetVirtualBoxMachine(virtualBox, szvmName, &machine)) {
		IConsole *console = NULL;
		IVRDEServerInfo *remoteDisplayInfo = NULL;
		BSTR guid;
		if (SUCCEEDED(machine->get_Id(&guid))) {
			if (SUCCEEDED(session->get_Console(&console))) {
				if (SUCCEEDED(console->get_VRDEServerInfo(&remoteDisplayInfo))) {
					remoteDisplayInfo->get_Port(&vrdpPort);
					SAFE_RELEASE(remoteDisplayInfo);
				}
				SAFE_RELEASE(console);
			}
			SAFE_RELEASE(machine);
		}
	}
	
	return vrdpPort;
}

DWORD GetMachineVRDPResolution()
{
	DWORD retval = 0;
	IMachine *machine = NULL;
	IDisplay *display = NULL;
	IConsole *console = NULL;
	ULONG displayWidth,displayHeight,bpp;
	
	if (GetVirtualBoxMachine(virtualBox, szvmName, &machine)) {
		BSTR guid;
		if (SUCCEEDED(machine->get_Id(&guid))) {
			if (SUCCEEDED(session->get_Console(&console)))	{
				if (SUCCEEDED(console->get_Display(&display))) {
					display->GetScreenResolution(0, &displayWidth, &displayHeight, &bpp);
					retval = displayWidth + (displayHeight << 16);
					SAFE_RELEASE(display);
				}
				SAFE_RELEASE(console);
			}
		}
		SAFE_RELEASE(machine);
	}
	
	return retval;
}

DWORD GetMachineState()
{
	DWORD retval = 0;
	IMachine *machine = NULL;
	MachineState state;
	if (GetVirtualBoxMachine(virtualBox, szvmName, &machine)) {
		if (SUCCEEDED(machine->get_State(&state))) {
			retval = state;
		}
		SAFE_RELEASE(machine);
	}
	
	return retval;
}

void Log(char loglevel, char* format, ...)
{
	if (loglevel > bLogLevel || bLogLevel == 0)
		return;
	va_list args;
	va_start(args, format);
	time_t rawtime;
	struct tm* timeinfo;
	char buf[80];

	time(&rawtime);
	timeinfo = localtime(&rawtime);
	FILE* flog;
	flog = fopen(szLogFile, "a");
	strftime(buf, 80, "%c", timeinfo);
	fprintf(flog, "[%s] ", buf);
	vfprintf(flog, format, args);
	fprintf(flog, "\n");
	va_end(args);
	fclose(flog);
}

void strcrpl(char *str, char c, char r)
{
	unsigned int i;
	if (str == NULL)
		return;
	for (i = 0; i < strlen(str); i++)
	{
		if (str[i] == c)
			str[i] = r;
	}
}
